# Copyright (c) 2022 Mario S. Könz; License: MIT
import typing as tp

__all__ = ["public_name"]


def public_name(cls: tp.Type[tp.Any], without_cls: bool = False) -> str:
    parts = []
    for part in cls.__module__.split("."):
        if part.startswith("_"):
            continue
        parts.append(part)

    public_module = "main"
    if parts:
        public_module = ".".join(parts)

    if without_cls:
        return f"{public_module}"
    return f"{public_module}.{cls.__name__}"



# Copyright (c) 2022 Mario S. Könz; License: MIT
import typing as tp

T = tp.TypeVar("T")  # pylint: disable=invalid-name


class StoreProtocol(tp.Protocol):
    def dump(self, obj: T) -> tp.Any:
        pass  # pragma: no cover

    def load(self, dataclass: type[T], **filter_kwgs: tp.Any) -> T:
        pass  # pragma: no cover



# Copyright (c) 2022 Mario S. Könz; License: MIT
import dataclasses as dc
import datetime
import enum
import types
import typing as tp

from django.db import models

from ._decorator import BACKEND_LINKER
from ._util import public_name

__all__ = ["CreateDjangoModel", "create_django_model"]


@dc.dataclass
class CreateDjangoModel:
    dataclass: type

    def __post_init__(self) -> None:
        # evaluate the Meta class
        pk_key, unique_together = self.extract_meta()
        # translate fields
        fields: dict[str, tp.Any] = self.translate_fields(pk_key)
        fields["Meta"] = self.generate_meta(unique_together)
        fields["__module__"] = fields["Meta"].app_label
        django_model = type(self.dataclass.__name__, (models.Model,), fields)
        BACKEND_LINKER.link(self.dataclass, django_model)

    def extract_meta(self) -> tuple[str | None, list[str] | None]:
        pk_key, unique_together = None, None
        if hasattr(self.dataclass, "Meta"):
            meta = self.dataclass.Meta  # type: ignore
            if hasattr(meta, "primary_key"):
                pk_key = meta.primary_key
            if hasattr(meta, "unique_together"):
                unique_together = meta.unique_together

        return pk_key, unique_together

    def translate_fields(
        self, pk_key: str | None
    ) -> "dict[str, models.Field[tp.Any, tp.Any]]":
        fields = {}
        for field in dc.fields(self.dataclass):
            django_field, opts = self.django_field_precursor(field.type)
            if field.name == pk_key:
                opts["primary_key"] = True
                if django_field is models.ForeignKey:
                    django_field = models.OneToOneField

            fields[field.name] = django_field(**opts)
        return fields

    def generate_meta(self, unique_together: list[str] | None) -> type:
        class Meta:
            app_label = public_name(self.dataclass, without_cls=True)
            db_table = public_name(self.dataclass)

        if unique_together:
            Meta.unique_together = unique_together  # type: ignore

        return Meta

    @classmethod
    def django_field_precursor(
        cls, type_: type
    ) -> "tuple[type[models.Field[tp.Any, tp.Any]], dict[str, tp.Any]]":
        # pylint: disable=too-many-return-statements,too-many-branches
        if type_ == str:
            return models.CharField, dict(max_length=1024)
        if type_ == int:
            return models.IntegerField, {}
        if type_ == float:
            return models.FloatField, {}
        if type_ == datetime.datetime:
            return models.DateTimeField, {}
        if type_ == datetime.date:
            return models.DateField, {}
        if type_ == datetime.time:
            return models.TimeField, {}
        if type_ == bytes:
            return models.BinaryField, dict(editable=True)

        if isinstance(type_, types.GenericAlias):
            origin = tp.get_origin(type_)
            subtypes = tp.get_args(type_)
            if origin is set:
                assert len(subtypes) == 1
                subtype = subtypes[0]
                try:
                    fk_class = BACKEND_LINKER.backend_class(subtype)
                    assert issubclass(fk_class, models.Model)
                    return models.ManyToManyField, dict(to=fk_class, related_name="+")
                except KeyError:
                    pass

        if isinstance(type_, types.UnionType):
            target_type, none_type = tp.get_args(type_)
            if none_type is type(None):
                field, kwgs = cls.django_field_precursor(target_type)
                kwgs["blank"] = True
                kwgs["null"] = True
                return field, kwgs

        if issubclass(type_, enum.Enum):
            max_length = 256
            choices = []
            for val in type_.__members__.values():
                choices.append((val.value, val.value))
                assert len(val.value) < max_length

            return models.CharField, dict(max_length=max_length, choices=choices)

        try:  # try Foreign Key relation (many-to-one)
            fk_class = BACKEND_LINKER.backend_class(type_)
            assert issubclass(fk_class, models.Model)
            return models.ForeignKey, dict(
                to=fk_class, on_delete=models.CASCADE, related_name="+"
            )

        except KeyError:
            pass

        raise NotImplementedError(type_)


def create_django_model(dataclass: type) -> type:
    CreateDjangoModel(dataclass)
    return dataclass



# Copyright (c) 2022 Mario S. Könz; License: MIT
import collections
import contextlib
import io
import typing as tp

import django
from django.apps import apps
from django.conf import settings
from django.core.management import call_command
from django.db import connections

from ._decorator import BACKEND_LINKER
from ._django_store import DjangoStore
from ._protocols import StoreProtocol

__all__ = ["set_store", "set_store_for_django", "ACTIVE_STORES"]

ACTIVE_STORES: collections.OrderedDict[str, StoreProtocol] = collections.OrderedDict()


def set_store(
    identifier: str = "first", flavor: str = "sqlite3", **kwgs: tp.Any
) -> None:
    if not settings.configured:
        settings.configure(
            DATABASES={"default": {"ENGINE": "django.db.backends.dummy"}},
            TIME_ZONE="UTC",
            USE_TZ=True,
            DEFAULT_AUTO_FIELD="django.db.models.BigAutoField",
        )
        django.setup()

    assert identifier != "default"
    databases = settings.DATABASES
    if flavor == "sqlite3":
        assert identifier not in databases
        databases[identifier] = dict(
            NAME=kwgs["path"],
            ENGINE="django.db.backends.sqlite3",
            TIME_ZONE="UTC",
            USE_TZ=True,
        )
        ACTIVE_STORES[identifier] = DjangoStore(identifier)
    else:
        raise NotImplementedError(flavor)

    apps.apps_ready = apps.models_ready = apps.loading = apps.ready = False
    apps.clear_cache()
    for app_label in BACKEND_LINKER.app_labels:
        if app_label not in settings.INSTALLED_APPS:
            settings.INSTALLED_APPS.append(app_label)

    apps.populate(settings.INSTALLED_APPS)
    connections.configure_settings(databases)

    call_command("makemigrations")
    with contextlib.redirect_stdout(io.StringIO()):
        call_command("migrate", database=identifier)


def set_store_for_django(identifier: str = "default") -> None:
    ACTIVE_STORES[identifier] = DjangoStore(identifier)



# Copyright (c) 2022 Mario S. Könz; License: MIT
import dataclasses as dc
import enum
import typing as tp

from django.db import models

from ._decorator import BACKEND_LINKER
from ._protocols import T


@dc.dataclass
class DjangoStore:
    identifier: str

    def dump(self, dc_obj: tp.Any) -> models.Model:
        django_model: type[models.Model] = BACKEND_LINKER.backend_class(dc_obj)
        kwgs, m2m = self.dataclass_to_django(dc_obj)
        dj_obj, created = django_model.objects.using(self.identifier).update_or_create(
            **kwgs
        )

        for key, vals in m2m.items():
            # remove old ones
            if not created:
                getattr(dj_obj, key).clear()
            for val in vals:
                getattr(dj_obj, key).add(self.dump(val))

        return dj_obj  # type: ignore

    def load(self, dataclass: type[T], **filter_kwgs: tp.Any) -> T:
        django_model: type[models.Model] = BACKEND_LINKER.backend_class(dataclass)
        instance = (
            django_model.objects.using(self.identifier).filter(**filter_kwgs).first()
        )
        return self.django_to_dataclass(instance)  # type: ignore

    def dataclass_to_django(self, dc_obj: models.Model) -> tp.Any:
        django_model: type[models.Model] = BACKEND_LINKER.backend_class(dc_obj)
        kwgs = {}
        m2m = {}
        for field in dc.fields(dc_obj):
            key = field.name
            val = getattr(dc_obj, key)
            if type(val) in BACKEND_LINKER.dc_to_backend:
                val = self.dump(val)
            if isinstance(val, enum.Enum):
                val = val.value

            # pylint: disable=protected-access
            dj_model = django_model._meta.get_field(key)
            if isinstance(dj_model, models.ManyToManyField):
                m2m[key] = val
            else:
                kwgs[key] = val

        return kwgs, m2m

    def django_to_dataclass(self, dj_obj: models.Model) -> tp.Any:
        dataclass = BACKEND_LINKER.backend_to_dc[type(dj_obj)]
        obj_kwgs = {}
        for field in dc.fields(dataclass):
            key = field.name
            val = getattr(dj_obj, key)
            if type(val) in BACKEND_LINKER.backend_to_dc:
                val = self.django_to_dataclass(val)
            field = [field for field in dc.fields(dataclass) if field.name == key][0]
            if issubclass(field.type, enum.Enum):
                val = field.type(val)
            # pylint: disable=protected-access
            if isinstance(dj_obj._meta.get_field(key), models.ManyToManyField):
                val = {self.django_to_dataclass(x) for x in val.all()}

            obj_kwgs[field.name] = val
        return dataclass(**obj_kwgs)
